/*
 * testChevaucheBeta.cpp
 *
 *  Created on: 10 nov. 2017
 *      Author: ddhoutau
 */

/*
 * testChevauche.cpp
 *
 *  Created on: 11 juil. 2017
 *      Author: ddhoutau
 */

#include <math.h>
#include <iostream>
#include <fstream>
#include <gmpxx.h>
#include <random>
#include <vector>
#include <map>
#include <assert.h>

#include <tclap/CmdLine.h>

using namespace std;

int main(int _argc, char **_argv) {

    int nbNodes;
    long int windowWidth, packetSize;
    int nbTirages;
    int nbRetrans;
    long int backoff;
    long int packetDuration;
    long int beta;
    long int tp;
    long int ts;

    try {
        TCLAP::CmdLine cmd("testChevaucheBeta", ' ', "1.0");
        TCLAP::ValueArg<int> nbNodesParam("n", "nbNodes", "nombre de noeuds", false, 3, "int");
        TCLAP::ValueArg<int> betaParam("b", "beta", "beta", false, 1000, "long int");
        TCLAP::ValueArg<int> tpParam("", "tp", "durée d'un pulse en fs", false, 100, "long int");
        TCLAP::ValueArg<long int> windowWidthParam("w", "windowWidth", "largeur de la fenêtre", false, 100, "long int");
        TCLAP::ValueArg<long int> packetSizeParam("p", "packetSize", "taille du paquet en bits", false, 10, "long int");
        TCLAP::ValueArg<int> nbTiragesParam("", "nbTirages", "nombre de tirages", false, 1000, "int");
        TCLAP::ValueArg<int> backoffParam("", "backoff", "backoff désiré", false, 100, "long int");

        cmd.add(nbNodesParam);
        cmd.add(betaParam);
        cmd.add(tpParam);
        cmd.add(windowWidthParam);
        cmd.add(packetSizeParam);
        cmd.add(nbTiragesParam);
        cmd.add(backoffParam);

        cmd.parse( _argc, _argv );

        nbNodes = nbNodesParam.getValue();
        beta = betaParam.getValue();
        tp = tpParam.getValue();
        windowWidth = windowWidthParam.getValue();
        packetSize = packetSizeParam.getValue();
        nbTirages = nbTiragesParam.getValue();
        backoff = backoffParam.getValue();

    } catch (TCLAP::ArgException &e) {
        cerr << "*** ERROR *** " << e.error() << " for arg " << e.argId() << std::endl;
    }

    packetDuration = packetSize*beta*tp;
    backoff = windowWidth - packetDuration;
    ts = tp * beta;
    std::random_device rd;
    std::mt19937 gen(rd());
    std::uniform_int_distribution<long int> disPulse(0, ts-tp);
    std::uniform_int_distribution<long int> disPacket(0, backoff);

    cout << "nbNodes:        " << nbNodes << endl;
    cout << "beta:           " << beta << endl;
    cout << "tp:             " << tp << endl;
    cout << "windowWidth:    " << windowWidth << endl;
    cout << "packetSize:     " << packetSize << endl;
    cout << "packetDuration: " << packetDuration << endl;
    cout << "backoff:        " << backoff << endl;
    cout << "nbTirages:      " << nbTirages << endl;

    vector<long int> vectMultiplexingDistribution(nbNodes+1,0);
    int nbConcurrent;
    bool fin;
    int n;
    multimap<long int, char> eventsMap;

    long int previousDate = 0;

    // tirage des impulsions

    for (int tirage=0; tirage<nbTirages; tirage++) {
        previousDate = 0;
        eventsMap.clear();

        for (n=0; n<nbNodes; n++) {
            long int date = disPulse(gen);
            assert(date>=0);
            //cout << "tirage:" << tirage << " pulse:" << n << "  debut:" << date << endl;
            eventsMap.insert(pair<long int, char>(date,'d'));
            eventsMap.insert(pair<long int, char>(date+packetDuration,'f'));
        }

    }

    // tirage des paquets

    for (int tirage=0; tirage<nbTirages; tirage++) {
        previousDate = 0;
        eventsMap.clear();

        for (n=0; n<nbNodes; n++) {
            long int date = disPacket(gen);
            assert(date>=0);
            //cout << "tirage:" << tirage << " packet:" << n << "  debut:" << date << endl;
            eventsMap.insert(pair<long int, char>(date,'d'));
            eventsMap.insert(pair<long int, char>(date+packetDuration,'f'));
        }

        nbConcurrent = 0;
        fin = false;
        n = 1;
        nbRetrans = 0;
        int nbFinished = 0;

//        for (auto it=eventsMap.begin(); it != eventsMap.end(); it++) {
//            cout << it->first << " " << it->second << endl;
//        }

        multimap<long int, char>::iterator first;

        long int elapsedTime=0;
        long int elapsedTimeTirage = 0;

        while (!eventsMap.empty()) {
            first = eventsMap.begin();
            elapsedTime = first->first - previousDate;
            vectMultiplexingDistribution[nbConcurrent] += elapsedTime;
            if (elapsedTime < 0) {
                cout << "previousDate: " << previousDate << endl;
                cout << "event date:   " << first->first << endl;
                cout << "elapsedTime:  " << elapsedTime << endl;
            }
            assert(elapsedTime >= 0);
            //cout << "et: " << elapsedTime << endl;
            if (first->second == 'd') {
                nbConcurrent++;
                //cout << "d " << first->first << "(" << nbConcurrent << ")" << endl;
            } else if ( first->second == 'f' ) {
                nbConcurrent--;
                assert(nbConcurrent >= 0);
                //cout << "f " << first->first << "(" << nbConcurrent << ")" << endl;
            }
            previousDate = first->first;
            eventsMap.erase(first);
            elapsedTimeTirage += elapsedTime;
        }

        elapsedTime = windowWidth - previousDate;
        //cout << "et: " << elapsedTime << endl;
        assert(nbConcurrent == 0);
        vectMultiplexingDistribution[nbConcurrent] += elapsedTime;
        elapsedTimeTirage += elapsedTime;

        //assert(elapsedTimeTirage == 100);
        //cout << "elapsedTimeTirage: " << elapsedTimeTirage << endl;

        /*
        while ( !fin ) {
            if (eventsMap.empty() || nbFinished >= redondance) {
                fin = true;
            } else {
                first = eventsMap.begin();
                if (first->second == 'd') {
                    nbRetrans++;
                    if ( nbRetrans == redondance ) {
                        dateRedondance += eventsMap.begin()->first;
                    }
                } else if (first->second == 'f') {
                    nbFinished++;
                }
                eventsMap.erase(first);
            }
        }

        */

        //cout << "tirage:" << tirage << " nbRetrans:" << nbRetrans << endl << endl;
        //vectRetransDistribution[nbRetrans]++;
    }

    cout << "distribution des retransmissions" << endl;
    long int totalTime = 0;
    for (int i=0; i<vectMultiplexingDistribution.size(); i++) {
        cout << i << ": " << (double)vectMultiplexingDistribution[i]/(nbTirages*windowWidth) << endl;
        totalTime += vectMultiplexingDistribution[i];
    }

    cout << "totalTime: " << totalTime << endl;
}







