#include <cstdio>
#include <cmath>
#include <random>
#include <iostream>
#include <algorithm>    // std::sort
#include <tclap/CmdLine.h>
#include <set>


int choose(int k, int n){
  if ( k > n ){
      std::cerr << " k can't be greater than n " << std::endl;
      exit(EXIT_FAILURE);
  }
  
  int a = 1; // n!
  int b; // k!
  int c; // (n-r)!
  for ( int i =1; i <= n; i++){
    a *= i;
    if ( i == k ){
      b = a; 
    }
    if ( i == n-k){
      c = a;
    }
  }
  
  int res;
  res = a/(b*c);
  return res;
  
}




int main (int _argc, char **_argv) {
  
  using namespace std;
  
  int n;
  long double tp;
  long double nbBits;
  int nbTest;
  int redundancy;
  int beta;
  int seed;
  bool math;
  
  
      try {
        TCLAP::CmdLine cmd("testMinBackoff", ' ', "1.0");
        TCLAP::ValueArg<int> nbNodesParam("n", "nbNodes", "nombre de noeuds", false, 1000, "int");
        TCLAP::ValueArg<long int> tpParam("", "tp", "pulse duration", false, 100, "long int");
        TCLAP::ValueArg<long int> packetSizeParam("p", "packetSize", "packet size (bit)", false, 1000, "long int");
        TCLAP::ValueArg<int> nbTestParam("", "nbTest", "nombre de test", false, 5000000, "int");
        TCLAP::ValueArg<int> redondanceParam("r", "redondance", "redondance désirée", false, 1, "int");
        TCLAP::ValueArg<int> betaParam("b", "beta", "beta utilisé", false, 1000, "int");
        TCLAP::ValueArg<int> seedParam("s", "seed", "RNG seed", false, 0, "int");
        TCLAP::SwitchArg mathModeParam ("m","math","mode math", cmd, false);

        
        cmd.add(nbNodesParam);
        cmd.add(tpParam);
        cmd.add(packetSizeParam);
        cmd.add(nbTestParam);
        cmd.add(redondanceParam);
        cmd.add(betaParam);
        cmd.add(seedParam);
        
        cmd.parse( _argc, _argv );

         n = nbNodesParam.getValue();
         tp = tpParam.getValue();
         beta = betaParam.getValue();
         nbBits = packetSizeParam.getValue();
         nbTest = nbTestParam.getValue();
         redundancy = redondanceParam.getValue();
         math = mathModeParam.getValue();
         seed = seedParam.getValue();
         
    } catch (TCLAP::ArgException &e) {
        cerr << "*** ERROR *** " << e.error() << " for arg " << e.argId() << std::endl;
    }

  
  
  
  
//   bool math = true;
  
  //   //les durée son en femtoseconde
  long double flight = 1666.6;
  
  
  long double winWindth = (2*((tp * beta * (nbBits-1) + tp) + flight )) ;
  
  if ( math ){
// // Here we consider the following : the minimum > to v means all the values > v    
// //       long double testedMin;  
// //       
// //       for ( testedMin = 0; testedMin <= winWindth; testedMin += 100000  ){
// //           long double proba;
// //           proba = pow((winWindth-testedMin)/(long double)winWindth,n);
// //           
// //     //       if ( proba != 0 ){
// //             cout << testedMin << "\t" << proba << endl;
// //     //       }
// //       }
    
      long double testedMin;  
      
      for ( testedMin = 1; testedMin <= winWindth; testedMin += 100000  ){
          long double proba,probaInf,probaSup,probaV;
          long int c;
          
          probaV= (winWindth-testedMin)/(long double)winWindth; // P(b>n)
          probaInf=pow(1-probaV,redundancy); //A: (P(b<v))^r proba que r valeurs < v
          probaSup=pow(probaV,n-redundancy); //B: (P(b>v))^(n-r) proba que le reste soient > v
          //c=    (redundancy,n);
          c=194280608456793000;
          proba = c*probaInf*probaSup; // P(A&B) = P(A)*P(B)
          
              
    //       if ( proba != 0 ){
            cout << testedMin << "\t" << proba << endl;
    //       }
            if (proba < 1.0e-100){ // Si la proba devien arbitrairement petite on arrète
              exit(0);
            }
      }
    
    
    
  }
  
  
  
  
  else { // Donne des résultats "empiriques"
 
      long double date = 0;
      mt19937_64 *RNG =  new mt19937_64(seed);
      uniform_real_distribution<float> distrib(0, winWindth);
//        cout << " winWidth " << winWindth;
      
      long double min =-1;
      long double accMin = 0;
      long double prevAccMin = 0;
      vector<long double> minList;

      redundancy--;
      if ( redundancy < 0 ) { cout << " Erreur redundancy " << endl; exit(1); }

      
      vector<long double> dateList;
        
      for ( int i = 0; i < nbTest ; i ++ ){
//         cout << " Test number " << i << endl;
        dateList.clear();
        
        for ( int node = 0; node < n; node ++){
          date = distrib(*RNG);
         
          dateList.push_back(date);   
//           dateList.insert(date);
          
          if ( dateList.size() > redundancy + 1 ){ 
            auto start = dateList.begin();
            auto end = dateList.end();  
            sort(start,end);
            dateList.pop_back();
          }
          
        }
        
//        accMin+=min;
        accMin+=dateList[redundancy];
//         minList.push_back(dateList[redundancy]);
        
        
        if ( accMin < prevAccMin ){
            cout << " ERROR accmin OVERFLOW " << endl;
            exit(1);
        }
        prevAccMin = accMin;
        
        
      }
      
      auto start = minList.begin();
      auto end = minList.end();  
      sort(start,end);
      for (auto it = end; it != start; it--){
          cout << *it << endl;
      }
      
      accMin/= nbTest;
      cout <<" "<<  accMin << " " << (accMin/winWindth)*100 <<"% ("  << (redundancy+1) << " " << n << ")"<< endl;
      
      
      
    }  

  return 0;
}